@@ -4,4 +4,4 @@ pub mod global_defs;
 pub mod hygiene;
 pub mod mangle_names;
 pub mod mangle_props;
-pub mod remove_parens;
+pub mod single;
